@@ -48,7 +48,7 @@
 /* for debugging purposes only */
 #ifdef DEBUG
 #undef DEBUG
-#define DEBUG(fmt,args...) { log_printf (LOG_ERROR,"%d: ",__LINE__); log_printf (LOG_ERROR,fmt,## args); }
+#define DEBUG(fmt,args...) { fprintf (stderr,"%d: ",__LINE__); fprintf (stderr,fmt,## args); }
 #else
 #undef DEBUG
 #define DEBUG(fmt,args...)
@@ -57,12 +57,8 @@
 #define KB(x) ((x) / 1024)
 #define PERCENTAGE(x,total) (((x) * 100) / (total))
 
-/* size of read/write buffer */
-#define BUFSIZE (10 * 1024)
-
 /* cmd-line flags */
 #define FLAG_NONE		0x00
-#define FLAG_VERBOSE	0x01
 #define FLAG_HELP		0x02
 #define FLAG_FILENAME	0x04
 #define FLAG_DEVICE		0x08
@@ -73,21 +69,34 @@
 #define LOG_NORMAL	1
 #define LOG_ERROR	2
 
-static void log_printf (int level,const char *fmt, ...)
+static NORETURN void log_failure (const char *fmt, ...)
 {
-	FILE *fp = level == LOG_NORMAL ? stdout : stderr;
 	va_list ap;
 	va_start (ap,fmt);
-	vfprintf (fp,fmt,ap);
+	vfprintf (stderr,fmt,ap);
 	va_end (ap);
-	fflush (fp);
+	fflush (stderr);
+
+	exit (EXIT_FAILURE);
+}
+
+static int verbose = 0;
+static void log_verbose (const char *fmt, ...)
+{
+	va_list ap;
+
+	if (!verbose)
+		return;
+
+	va_start (ap,fmt);
+	vfprintf (stdout,fmt,ap);
+	va_end (ap);
+	fflush (stdout);
 }
 
 static NORETURN void showusage(bool error)
 {
-	int level = error ? LOG_ERROR : LOG_NORMAL;
-
-	log_printf (level,
+	fprintf (error ? stderr : stdout,
 			"\n"
 			"Flash Copy - Written by Abraham van der Merwe <abraham@2d3d.co.za>\n"
 			"\n"
@@ -101,7 +110,7 @@ static NORETURN void showusage(bool error)
 			"   -A | --erase-all Erases the whole device regardless of the image size\n"
 			"   -V | --version   Show version information and exit\n"
 			"   <filename>       File which you want to copy to flash\n"
-			"   <device>         Flash device to write to (e.g. /dev/mtd0, /dev/mtd1, etc.)\n"
+			"   <device>         Flash device node or 'mtd:<name>' to write to (e.g. /dev/mtd0, /dev/mtd1, mtd:data, etc.)\n"
 			"\n",
 			PROGRAM_NAME);
 
@@ -110,49 +119,85 @@ static NORETURN void showusage(bool error)
 
 static int safe_open (const char *pathname,int flags)
 {
+	const char *access = "unknown";
 	int fd;
 
 	fd = open (pathname,flags);
 	if (fd < 0)
 	{
-		log_printf (LOG_ERROR,"While trying to open %s",pathname);
 		if (flags & O_RDWR)
-			log_printf (LOG_ERROR," for read/write access");
+			access = "read/write";
 		else if (flags & O_RDONLY)
-			log_printf (LOG_ERROR," for read access");
+			access = "read";
 		else if (flags & O_WRONLY)
-			log_printf (LOG_ERROR," for write access");
-		log_printf (LOG_ERROR,": %m\n");
-		exit (EXIT_FAILURE);
+			access = "write";
+
+		log_failure ("While trying to open %s for %s access: %m\n",pathname,access);
 	}
 
 	return (fd);
 }
 
-static void safe_read (int fd,const char *filename,void *buf,size_t count,bool verbose)
+static void safe_read (int fd,const char *filename,void *buf,size_t count)
 {
 	ssize_t result;
 
 	result = read (fd,buf,count);
 	if (count != result)
 	{
-		if (verbose) log_printf (LOG_NORMAL,"\n");
+		log_verbose ("\n");
 		if (result < 0)
 		{
-			log_printf (LOG_ERROR,"While reading data from %s: %m\n",filename);
-			exit (EXIT_FAILURE);
+			log_failure("While reading data from %s: %m\n",filename);
 		}
-		log_printf (LOG_ERROR,"Short read count returned while reading from %s\n",filename);
-		exit (EXIT_FAILURE);
+		log_failure("Short read count returned while reading from %s\n",filename);
 	}
 }
 
+static void safe_write (int fd,const void *buf,size_t count,size_t written,unsigned long long to_write,const char *device)
+{
+	ssize_t result;
+
+	/* write to device */
+	result = write (fd,buf,count);
+	if (count != result)
+	{
+		log_verbose ("\n");
+		if (result < 0)
+		{
+			log_failure("While writing data to 0x%.8lx-0x%.8lx on %s: %m\n",
+					written,written + count,device);
+		}
+		log_failure("Short write count returned while writing to x%.8zx-0x%.8zx on %s: %zu/%llu bytes written to flash\n",
+				written,written + count,device,written + result,to_write);
+	}
+}
+
+static off_t safe_lseek (int fd,off_t offset,int whence,const char *filename)
+{
+	off_t off;
+
+	off = lseek (fd,offset,whence);
+	if (off < 0)
+	{
+		log_failure("While seeking on %s: %m\n",filename);
+	}
+
+	return off;
+}
+
 static void safe_rewind (int fd,const char *filename)
 {
-	if (lseek (fd,0L,SEEK_SET) < 0)
+	safe_lseek(fd,0L,SEEK_SET,filename);
+}
+
+static void safe_memerase (int fd,const char *device,struct erase_info_user *erase)
+{
+	if (ioctl (fd,MEMERASE,erase) < 0)
 	{
-		log_printf (LOG_ERROR,"While seeking to start of %s: %m\n",filename);
-		exit (EXIT_FAILURE);
+		log_verbose ("\n");
+		log_failure("While erasing blocks 0x%.8x-0x%.8x on %s: %m\n",
+				(unsigned int) erase->start,(unsigned int) (erase->start + erase->length),device);
 	}
 }
 
@@ -170,12 +215,11 @@ int main (int argc,char *argv[])
 {
 	const char *filename = NULL,*device = NULL;
 	int i,flags = FLAG_NONE;
-	ssize_t result;
 	size_t size,written;
 	struct mtd_info_user mtd;
 	struct erase_info_user erase;
 	struct stat filestat;
-	unsigned char src[BUFSIZE],dest[BUFSIZE];
+	unsigned char *src,*dest;
 
 	/*********************
 	 * parse cmd-line
@@ -205,7 +249,7 @@ int main (int argc,char *argv[])
 				DEBUG("Got FLAG_HELP\n");
 				break;
 			case 'v':
-				flags |= FLAG_VERBOSE;
+				verbose = 1;
 				DEBUG("Got FLAG_VERBOSE\n");
 				break;
 			case 'p':
@@ -231,13 +275,19 @@ int main (int argc,char *argv[])
 		DEBUG("Got filename: %s\n",filename);
 
 		flags |= FLAG_DEVICE;
-		device = argv[optind+1];
+		device = mtd_find_dev_node(argv[optind+1]);
+		if (!device)
+			log_failure("Failed to find device %s\n", argv[optind+1]);
+
 		DEBUG("Got device: %s\n",device);
 	}
 
 	if (flags & FLAG_HELP || device == NULL)
 		showusage(flags != FLAG_HELP);
 
+	if (flags & FLAG_PARTITION && flags & FLAG_ERASE_ALL)
+		log_failure("Option --partition does not support --erase-all\n");
+
 	atexit (cleanup);
 
 	/* get some info about the flash device */
@@ -245,24 +295,25 @@ int main (int argc,char *argv[])
 	if (ioctl (dev_fd,MEMGETINFO,&mtd) < 0)
 	{
 		DEBUG("ioctl(): %m\n");
-		log_printf (LOG_ERROR,"This doesn't seem to be a valid MTD flash device!\n");
-		exit (EXIT_FAILURE);
+		log_failure("This doesn't seem to be a valid MTD flash device!\n");
 	}
 
 	/* get some info about the file we want to copy */
 	fil_fd = safe_open (filename,O_RDONLY);
 	if (fstat (fil_fd,&filestat) < 0)
-	{
-		log_printf (LOG_ERROR,"While trying to get the file status of %s: %m\n",filename);
-		exit (EXIT_FAILURE);
-	}
+		log_failure("While trying to get the file status of %s: %m\n",filename);
 
 	/* does it fit into the device/partition? */
 	if (filestat.st_size > mtd.size)
-	{
-		log_printf (LOG_ERROR,"%s won't fit into %s!\n",filename,device);
-		exit (EXIT_FAILURE);
-	}
+		log_failure("%s won't fit into %s!\n",filename,device);
+
+	src = malloc(mtd.erasesize);
+	if (!src)
+		log_failure("Malloc failed");
+
+	dest = malloc(mtd.erasesize);
+	if (!dest)
+		log_failure("Malloc failed");
 
 	/* diff block flashcp */
 	if (flags & FLAG_PARTITION)
@@ -288,37 +339,24 @@ int main (int argc,char *argv[])
 		erase.length *= mtd.erasesize;
 	}
 
-	if (flags & FLAG_VERBOSE)
+	if (verbose)
 	{
 		/* if the user wants verbose output, erase 1 block at a time and show him/her what's going on */
 		int blocks = erase.length / mtd.erasesize;
 		erase.length = mtd.erasesize;
-		log_printf (LOG_NORMAL,"Erasing blocks: 0/%d (0%%)",blocks);
+		log_verbose ("Erasing blocks: 0/%d (0%%)",blocks);
 		for (i = 1; i <= blocks; i++)
 		{
-			log_printf (LOG_NORMAL,"\rErasing blocks: %d/%d (%d%%)",i,blocks,PERCENTAGE (i,blocks));
-			if (ioctl (dev_fd,MEMERASE,&erase) < 0)
-			{
-				log_printf (LOG_NORMAL,"\n");
-				log_printf (LOG_ERROR,
-						"While erasing blocks 0x%.8x-0x%.8x on %s: %m\n",
-						(unsigned int) erase.start,(unsigned int) (erase.start + erase.length),device);
-				exit (EXIT_FAILURE);
-			}
+			log_verbose ("\rErasing blocks: %d/%d (%d%%)",i,blocks,PERCENTAGE (i,blocks));
+			safe_memerase(dev_fd,device,&erase);
 			erase.start += mtd.erasesize;
 		}
-		log_printf (LOG_NORMAL,"\rErasing blocks: %d/%d (100%%)\n",blocks,blocks);
+		log_verbose ("\rErasing blocks: %d/%d (100%%)\n",blocks,blocks);
 	}
 	else
 	{
 		/* if not, erase the whole chunk in one shot */
-		if (ioctl (dev_fd,MEMERASE,&erase) < 0)
-		{
-			log_printf (LOG_ERROR,
-					"While erasing blocks from 0x%.8x-0x%.8x on %s: %m\n",
-					(unsigned int) erase.start,(unsigned int) (erase.start + erase.length),device);
-			exit (EXIT_FAILURE);
-		}
+		safe_memerase(dev_fd,device,&erase);
 	}
 	DEBUG("Erased %u / %luk bytes\n",erase.length,filestat.st_size);
 
@@ -326,48 +364,30 @@ int main (int argc,char *argv[])
 	 * write the entire file to flash *
 	 **********************************/
 
-	if (flags & FLAG_VERBOSE) log_printf (LOG_NORMAL,"Writing data: 0k/%lluk (0%%)",KB ((unsigned long long)filestat.st_size));
+	log_verbose ("Writing data: 0k/%lluk (0%%)",KB ((unsigned long long)filestat.st_size));
 	size = filestat.st_size;
-	i = BUFSIZE;
+	i = mtd.erasesize;
 	written = 0;
 	while (size)
 	{
-		if (size < BUFSIZE) i = size;
-		if (flags & FLAG_VERBOSE)
-			log_printf (LOG_NORMAL,"\rWriting data: %dk/%lluk (%llu%%)",
-					KB (written + i),
-					KB ((unsigned long long)filestat.st_size),
-					PERCENTAGE (written + i,(unsigned long long)filestat.st_size));
+		if (size < mtd.erasesize) i = size;
+		log_verbose ("\rWriting data: %dk/%lluk (%llu%%)",
+				KB (written + i),
+				KB ((unsigned long long)filestat.st_size),
+				PERCENTAGE ((unsigned long long)written + i,(unsigned long long)filestat.st_size));
 
 		/* read from filename */
-		safe_read (fil_fd,filename,src,i,flags & FLAG_VERBOSE);
+		safe_read (fil_fd,filename,src,i);
 
 		/* write to device */
-		result = write (dev_fd,src,i);
-		if (i != result)
-		{
-			if (flags & FLAG_VERBOSE) log_printf (LOG_NORMAL,"\n");
-			if (result < 0)
-			{
-				log_printf (LOG_ERROR,
-						"While writing data to 0x%.8lx-0x%.8lx on %s: %m\n",
-						written,written + i,device);
-				exit (EXIT_FAILURE);
-			}
-			log_printf (LOG_ERROR,
-					"Short write count returned while writing to x%.8zx-0x%.8zx on %s: %zu/%llu bytes written to flash\n",
-					written,written + i,device,written + result,(unsigned long long)filestat.st_size);
-			exit (EXIT_FAILURE);
-		}
+		safe_write(dev_fd,src,i,written,(unsigned long long)filestat.st_size,device);
 
 		written += i;
 		size -= i;
 	}
-	if (flags & FLAG_VERBOSE)
-		log_printf (LOG_NORMAL,
-				"\rWriting data: %lluk/%lluk (100%%)\n",
-				KB ((unsigned long long)filestat.st_size),
-				KB ((unsigned long long)filestat.st_size));
+	log_verbose ("\rWriting data: %lluk/%lluk (100%%)\n",
+			KB ((unsigned long long)filestat.st_size),
+			KB ((unsigned long long)filestat.st_size));
 	DEBUG("Wrote %d / %lluk bytes\n",written,(unsigned long long)filestat.st_size);
 
 	/**********************************
@@ -377,42 +397,34 @@ int main (int argc,char *argv[])
 	safe_rewind (fil_fd,filename);
 	safe_rewind (dev_fd,device);
 	size = filestat.st_size;
-	i = BUFSIZE;
+	i = mtd.erasesize;
 	written = 0;
-	if (flags & FLAG_VERBOSE) log_printf (LOG_NORMAL,"Verifying data: 0k/%lluk (0%%)",KB ((unsigned long long)filestat.st_size));
+	log_verbose ("Verifying data: 0k/%lluk (0%%)",KB ((unsigned long long)filestat.st_size));
 	while (size)
 	{
-		if (size < BUFSIZE) i = size;
-		if (flags & FLAG_VERBOSE)
-			log_printf (LOG_NORMAL,
-					"\rVerifying data: %luk/%lluk (%llu%%)",
-					KB (written + i),
-					KB ((unsigned long long)filestat.st_size),
-					PERCENTAGE (written + i,(unsigned long long)filestat.st_size));
+		if (size < mtd.erasesize) i = size;
+		log_verbose ("\rVerifying data: %luk/%lluk (%llu%%)",
+				KB (written + i),
+				KB ((unsigned long long)filestat.st_size),
+				PERCENTAGE ((unsigned long long)written + i,(unsigned long long)filestat.st_size));
 
 		/* read from filename */
-		safe_read (fil_fd,filename,src,i,flags & FLAG_VERBOSE);
+		safe_read (fil_fd,filename,src,i);
 
 		/* read from device */
-		safe_read (dev_fd,device,dest,i,flags & FLAG_VERBOSE);
+		safe_read (dev_fd,device,dest,i);
 
 		/* compare buffers */
 		if (memcmp (src,dest,i))
-		{
-			log_printf (LOG_ERROR,
-					"File does not seem to match flash data. First mismatch at 0x%.8zx-0x%.8zx\n",
+			log_failure("File does not seem to match flash data. First mismatch at 0x%.8zx-0x%.8zx\n",
 					written,written + i);
-			exit (EXIT_FAILURE);
-		}
 
 		written += i;
 		size -= i;
 	}
-	if (flags & FLAG_VERBOSE)
-		log_printf (LOG_NORMAL,
-				"\rVerifying data: %lluk/%lluk (100%%)\n",
-				KB ((unsigned long long)filestat.st_size),
-				KB ((unsigned long long)filestat.st_size));
+	log_verbose ("\rVerifying data: %lluk/%lluk (100%%)\n",
+			KB ((unsigned long long)filestat.st_size),
+			KB ((unsigned long long)filestat.st_size));
 	DEBUG("Verified %d / %lluk bytes\n",written,(unsigned long long)filestat.st_size);
 
 	exit (EXIT_SUCCESS);
@@ -434,56 +446,39 @@ DIFF_BLOCKS:
 	int blocks = erase.length / mtd.erasesize;
 	erase.length = mtd.erasesize;
 
-	if (flags & FLAG_VERBOSE)
-		log_printf (LOG_NORMAL,
-				"\rProcessing blocks: 0/%d (%d%%)", blocks, PERCENTAGE (0,blocks));
+	log_verbose ("\rProcessing blocks: 0/%d (%d%%)", blocks, PERCENTAGE (0,blocks));
 	for (int s = 1; s <= blocks; s++)
 	{
 		if (size < mtd.erasesize) i = size;
-		if (flags & FLAG_VERBOSE)
-			log_printf (LOG_NORMAL,
-					"\rProcessing blocks: %d/%d (%d%%)", s, blocks, PERCENTAGE (s,blocks));
+		log_verbose ("\rProcessing blocks: %d/%d (%d%%)", s, blocks, PERCENTAGE (s,blocks));
 
 		/* read from filename */
-		safe_read (fil_fd,filename,src,i,flags & FLAG_VERBOSE);
+		safe_read (fil_fd,filename,src,i);
 
 		/* read from device */
-		current_dev_block = lseek(dev_fd, 0, SEEK_CUR);
-		safe_read (dev_fd,device,dest,i,flags & FLAG_VERBOSE);
+		current_dev_block = safe_lseek(dev_fd, 0, SEEK_CUR, device);
+		safe_read (dev_fd,device,dest,i);
 
 		/* compare buffers, if not the same, erase and write the block */
 		if (memcmp (src,dest,i))
 		{
 			diffBlock++;
 			/* erase block */
-			lseek(dev_fd, current_dev_block, SEEK_SET);
-			if (ioctl (dev_fd,MEMERASE,&erase) < 0)
-			{
-				log_printf (LOG_NORMAL,"\n");
-				log_printf (LOG_ERROR,
-						"While erasing blocks 0x%.8x-0x%.8x on %s: %m\n",
-						(unsigned int) erase.start,(unsigned int) (erase.start + erase.length),device);
-				exit (EXIT_FAILURE);
-			}
+			safe_lseek(dev_fd, current_dev_block, SEEK_SET, device);
+			safe_memerase(dev_fd,device,&erase);
 
 			/* write to device */
-			lseek(dev_fd, current_dev_block, SEEK_SET);
-			result = write (dev_fd,src,i);
-			if (i != result)
-			{
-				if (flags & FLAG_VERBOSE) log_printf (LOG_NORMAL,"\n");
-				if (result < 0)
-				{
-					log_printf (LOG_ERROR,
-							"While writing data to 0x%.8lx-0x%.8lx on %s: %m\n",
-							written,written + i,device);
-					exit (EXIT_FAILURE);
-				}
-				log_printf (LOG_ERROR,
-						"Short write count returned while writing to x%.8zx-0x%.8zx on %s: %zu/%llu bytes written to flash\n",
-						written,written + i,device,written + result,(unsigned long long)filestat.st_size);
-				exit (EXIT_FAILURE);
-			}
+			safe_lseek(dev_fd, current_dev_block, SEEK_SET, device);
+			safe_write(dev_fd,src,i,written,(unsigned long long)filestat.st_size,device);
+
+			/* read from device */
+			safe_lseek(dev_fd, current_dev_block, SEEK_SET, device);
+			safe_read (dev_fd,device,dest,i);
+
+			/* compare buffers for write success */
+			if (memcmp (src,dest,i))
+				log_failure("File does not seem to match flash data. First mismatch at 0x%.8zx-0x%.8zx\n",
+						written,written + i);
 		}
 
 		erase.start += i;
@@ -491,7 +486,7 @@ DIFF_BLOCKS:
 		size -= i;
 	}
 
-	if (flags & FLAG_VERBOSE) log_printf (LOG_NORMAL, "\ndiff blocks: %d\n", diffBlock);
+	log_verbose ("\ndiff blocks: %d\n", diffBlock);
 
 	exit (EXIT_SUCCESS);
 }
